
//----------------------------------------------------------------------------
// StarChase -- an interstellar action extravaganza 
// by Christopher Rasmussen
// v 1.1 (c) 2010
//----------------------------------------------------------------------------

#include <iostream>
#include <cstdlib>
#include <time.h>
#include <ncurses.h>

#include "rand_utils.hh"

using namespace std;

//----------------------------------------------------------------------------

const int SCREEN_WIDTH     = 20;
const int SCREEN_HEIGHT    = 15;
const int NUM_STARS        = 20;
const int OSCILLATOR_INDEX = 0;

const char FORWARD_KEY     = 'w';
const char REVERSE_KEY     = 's';
const char LEFT_KEY        = 'a';
const char RIGHT_KEY       = 'd';

// color 0 is reserved

const int SHIP_COLOR       = 1;
const int STAR_COLOR       = 2;
const int OSCILLATOR_COLOR = 3;

// global variable alert!!

bool using_color = false;

//----------------------------------------------------------------------------

bool in_boundaries(int, int);

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

class Star
{
public:

  Star();
  void draw(bool);
  void wiggle();     // move randomly

  int x, y;          // position
  bool collected;    // has ship gotten it yet?
};

//----------------------------------------------------------------------------

class Ship
{
public:

  Ship();
  void draw();
  void keyboard_control(char);
  void check_for_star_collections(Star *, int);

  int x, y;                  // position
  int dx, dy;                // direction
  int num_stars_collected;   // how many stars user has so far

};

//----------------------------------------------------------------------------

class Score
{
public:

  Score();
  void draw();
  void goodbye_message();
  void update_time();
  bool enough_stars(int);

  int time_elapsed;
  time_t start_time, now_time;
  int num_stars_collected;
};

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// distribute stars randomly and say none have been collected yet
// some may be on top of each other

Star::Star()
{
  x = rand_low_high(1, SCREEN_WIDTH - 2);
  y = rand_low_high(1, SCREEN_HEIGHT - 2);
  collected = false;
}

//----------------------------------------------------------------------------

void Star::draw(bool is_oscillator)
{
  if (!collected) {
    move(y, x);
    if (is_oscillator)
      printw("X");
    else
      printw("*");
  }
}

//----------------------------------------------------------------------------

// move star to random neighbor location (or stay put), as long as it 
// doesn't leave the screen

void Star::wiggle()
{
  int rand_dx = rand_low_high(-1, 1);
  int rand_dy = rand_low_high(-1, 1);
  
  if (in_boundaries(x + rand_dx, y + rand_dy)) {
    x += rand_dx;
    y += rand_dy;
  }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

Ship::Ship()
{
  x = SCREEN_WIDTH/2;
  y = SCREEN_HEIGHT/2;
  dx = 0;
  dy = -1;
  num_stars_collected = 0;
}

//----------------------------------------------------------------------------

// show ship with current location and heading

void Ship::draw()
{
  if (using_color) 
    attron(COLOR_PAIR(SHIP_COLOR));  

  move(y, x);
  printw("0");
  move(y + dy, x + dx);
  printw("+");

  if (using_color) 
    attroff(COLOR_PAIR(SHIP_COLOR));
}

//----------------------------------------------------------------------------

// loop over all uncollected stars and see which ones we're on top of.
// change collection status of any such stars and return how many we got.

void Ship::check_for_star_collections(Star *stars, int num_stars)
{
  for (int i = 0; i < num_stars; i++) 
    if (!stars[i].collected && x == stars[i].x && y == stars[i].y) {
      stars[i].collected = true;
      num_stars_collected++;
    }
}

//----------------------------------------------------------------------------

// based on key input, move ship as commanded

void Ship::keyboard_control(char c)
{
  int temp;

  switch (c) {
    
    // move forward				
      
  case FORWARD_KEY:
    if (in_boundaries(x + dx, y + dy)) {
      x += dx;
      y += dy;
    }
    break;
    
    // move backward
    
  case REVERSE_KEY:
    if (in_boundaries(x - dx, y - dy)) {
      x -= dx;
      y -= dy;
    }
    break;

    // rotate left 90 degs.

  case LEFT_KEY:
    temp = -dx;
    dx = dy;
    dy = temp;
    break;

    // rotate right 90 degs.

  case RIGHT_KEY:
    temp = dx;
    dx = -dy;
    dy = temp;
    break;
  }

}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

Score::Score()
{
  start_time = time(NULL);
}

//----------------------------------------------------------------------------

// how long has it been since beginning of game?

void Score::update_time()
{   
  now_time = time(NULL);
  time_elapsed = static_cast<int>(now_time - start_time);
}

//----------------------------------------------------------------------------

// draw user's score and time elapsed

void Score::draw()
{
  move(SCREEN_HEIGHT, 0);
  printw("%3i s  %2i/%i stars", time_elapsed, num_stars_collected, NUM_STARS);

}

//----------------------------------------------------------------------------

// keep track of how many stars have been collected.  return true if
// they're all gone

bool Score::enough_stars(int n)
{
  num_stars_collected = n;   // keep a copy for drawing
  return (num_stars_collected >= NUM_STARS);
}

//----------------------------------------------------------------------------

void Score::goodbye_message()
{
  if (num_stars_collected == NUM_STARS) {
    cout << "You got all of the stars in " << time_elapsed << " seconds\n";
    if (time_elapsed < 20) 
      cout << "Amazing!!\n";
    else if (time_elapsed < 25)
      cout << "Great!\n";
    else if (time_elapsed < 30)
      cout << "Nice job!\n";
    else if (time_elapsed < 35)
      cout << "Pretty good.\n";
    else if (time_elapsed < 40)
      cout << "OK...\n";
    else 
      cout << "Pfft...you need to step it up\n";
  }
  else
    cout << "Thanks for playing...\n";
}

//----------------------------------------------------------------------------

// this stuff probably should go in a header

void init_ncurses();

Star *initialize_stars(int);
void draw_stars(Star *, int);
void draw_boundaries();

//----------------------------------------------------------------------------

// get ready to draw with ncurses library

void init_ncurses()
{
  initscr();

  // comment this out to default to no colors
  
  if (has_colors()) {
    using_color = true;
    start_color();
    init_pair(SHIP_COLOR, COLOR_RED, COLOR_BLACK);
  }

  raw();
  noecho();
  curs_set(0);   // make cursor invisible
}

//----------------------------------------------------------------------------


Star *initialize_stars(int num_stars)
{
  return new Star[NUM_STARS];
}

//----------------------------------------------------------------------------

// outline edges of playing area

void draw_boundaries()
{
  int i;

  // top and bottom

  for (i = 0; i < SCREEN_WIDTH; i++) {
    move(0, i);
    printw(".");
    move(SCREEN_HEIGHT - 1, i);
    printw(".");
  }

  // left and right

  for (i = 0; i < SCREEN_HEIGHT; i++) {
    move(i, 0);
    printw(".");
    move(i, SCREEN_WIDTH - 1);
    printw(".");
  }
 
}

//----------------------------------------------------------------------------

// show all remaining stars

void draw_stars(Star *stars, int n)
{
  for (int i = 0; i < n; i++)
    stars[i].draw(i == OSCILLATOR_INDEX);
}

//----------------------------------------------------------------------------

// check that we are not touching the "edge of space"

bool in_boundaries(int x, int y)
{
  return (x >= 1 && x < (SCREEN_WIDTH - 1) &&
	  y >= 1 && y < (SCREEN_HEIGHT - 1));
}

//----------------------------------------------------------------------------

// print title, instructions

void draw_splash()
{
  printw("Welcome to StarChase, the wacky game of energy collection!!\n");
  printw("Try to visit all the stars as fast as possible...\n");
  printw("The X is an oscillating star...it's not dangerous--just a little harder to get\n");
  printw("%c/%c will move your ship forward/backward; %c/%c will turn your ship left/right\n", FORWARD_KEY, REVERSE_KEY, LEFT_KEY, RIGHT_KEY);   
  printw("Hit 'e' to exit at any time\n");   
  printw("Ready?\n");   
}

//----------------------------------------------------------------------------

int main(void)
{
  int c;                                  // hold user's latest keyboard input

  Star *star_array;                       // locations and statuses of all stars
  Ship ship;                              // initializes ship position and direction

  // get ready...

  init_rand();
  init_ncurses();

  draw_splash();
  refresh();			                                              
  c = getch();			                                              
  
  // set initial star positions and stats

  star_array = initialize_stars(NUM_STARS);

  Score score;  // clock starts in constructor
 
  // need this for motion without input

  timeout(100);

  // loop to allow ship motion

  while (c != 'e') {

    score.update_time();

    // one star oscillates

    star_array[OSCILLATOR_INDEX].wiggle();

    // clear screen & print it in current location

    erase();

    draw_boundaries();
    draw_stars(star_array, NUM_STARS);
    ship.draw();
    score.draw();

    refresh();

    // move ship

    c = getch();
    ship.keyboard_control(c);

    // check for star collection and update stats

    ship.check_for_star_collections(star_array, NUM_STARS);
    
    if (score.enough_stars(ship.num_stars_collected))
      break;

  }

  // get out of ncurses mode so terminal is not screwed up when program exits

  endwin();

  // parting words...

  score.goodbye_message();

  exit(1);	
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
